.val-box {
    font-size: 2rem;
    font-weight: bold;
    background-color: #cce5ff;
    border-color: #b8daff;
    border-radius: 6px;
    border-style: solid;
    border-width: 2px;
    text-align: center;
    display: block;
}

.min-val {
    color: gray;
    font-weight: 100;
}

.max-val {
    color: gray;
    font-weight: 100;
}

div.page-header {
    border-bottom: #2484e7;
    border-bottom-style: solid;
    margin-bottom: 20px;
    /*background-color: #a6ed97;*/

}

div.page-footer {
    border-top: black;
    border-top-style: solid;
    border-width: 1px;
    margin-top: 20px;
    padding-top: 20px;
    border-color: #2484e7;
}